import os
import time
from typing import List, Tuple
from dataclasses import dataclass
from decimal import Decimal
import math

import tiktoken
import replicate

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, get_tokens_encoded, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt


class ReplicateModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.max_retries = 5
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024 if model.max_output_tokens is None else model.max_output_tokens
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer

    def _call_api(self, prompt: str) -> str:

        response = replicate.run(self.spec.internal_name, input={
            "prompt": prompt,
            "system_prompt": self.spec.system_message if self.spec.system_message is not None else "",
            "top_k": 50,
            "top_p": 0.9,
            "temperature": 0,
            "max_new_tokens": 1024,
            "presence_penalty": 0,
            "frequency_penalty": 0
        })
        answer = "".join(response)
        return answer

    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, num_tokens_input = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        response = self._call_api(final_prompt)
        tokens_response = len(get_tokens_encoded(response, self.encoding))
        cost_input = (int(num_tokens_input) * Decimal(self.spec.price_per_1k_tokens / 1000)) if self.spec.price_per_1k_tokens is not None else Decimal(0)
        cost_output = (int(tokens_response) * Decimal(self.spec.price_per_1k_output_tokens / 1000)) if self.spec.price_per_1k_output_tokens is not None else Decimal(0)
        final_cost = cost_input + cost_output
        return response, final_cost



from typing import *

from parsee.datasets.dataset_dataclasses import DatasetRow
from parsee.extraction.extractor_dataclasses import Base64Image


class Prompt:

    def __init__(self, intro: Optional[str], main_task: str, additional_info: Optional[str] = None, full_example: Optional[str] = None, available_data: Optional[Union[str, List[Base64Image]]] = None,
                 history: Optional[List[str]] = None):
        self.intro = f"{intro} \n" if intro is not None else ""
        self.main_task = main_task
        self.additional_info = f"{additional_info} \n" if additional_info is not None else ""
        self.full_example = f"{full_example} \n" if full_example is not None else ""
        self.available_data = available_data if available_data is not None else ""
        self.history = ""
        if history is not None and len(history) > 0:
            self.history = "[PREVIOUS MESSAGES]\n"
            self.history += "\n".join(history) + "\n [END PREVIOUS MESSAGES]\n"

    def __str__(self) -> str:
        return f'''{self.history} {self.instructions()} {self.available_data_string()}'''

    def instructions(self) -> str:
        return f"""{self.intro} {self.main_task} \n {self.additional_info} {self.full_example}"""

    def available_data_string(self) -> str:
        return self.available_data if isinstance(self.available_data, str) else ''



from typing import *
from parsee.chat.custom_dataclasses import Message, ChatSettings
from decimal import Decimal
from parsee.storage.interfaces import DocumentManager
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.model_loader import get_llm_base_model
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.utils.helper import merge_answer_pieces


def run_chat(message: Message, message_history: List[Message], document_manager: DocumentManager, receivers: List[MlModelSpecification], most_recent_references_only: bool, show_chunk_index: bool = False, single_page_processing_max_images_trigger: Optional[int] = None) -> List[Message]:

    output = []

    models = [get_llm_base_model(spec) for spec in receivers]

    # collect all references if requested
    references = message.references if most_recent_references_only else []
    if not most_recent_references_only:
        added_references = set()
        all_messages = [message] + message_history
        for m in all_messages:
            new_references = [x for x in m.references if x.reference_id() not in added_references]
            for ref in new_references:
                references.append(ref)
                added_references.add(ref.reference_id())

    for model in models:
        data = document_manager.load_documents(references, model.spec.multimodal, str(message), model.spec.max_images, model.spec.max_tokens-document_manager.settings.min_tokens_for_instructions_and_history, show_chunk_index)
        # for multimodal queries, check if pages have to be processed individually
        process_pages_individually = False
        if type(data) is list:
            # check if pages can be processed one by one
            if single_page_processing_max_images_trigger is not None and len(data) >= single_page_processing_max_images_trigger:
                process_pages_individually = True
        if process_pages_individually:
            answers = []
            cost = Decimal(0)
            for k, img in enumerate(data):
                if k > 0:
                    additional_info = f"We are showing you the images contained in the document one by one. The current image is number {k + 1} out of a total of {len(data)}.\n Your last answer ended with the following (make sure that your new answer is valid JSON or similar, as requested; last 500 characters are shown):\n" \
                                      f"{answers[-1][-500:]}"
                else:
                    additional_info = f"We are showing you the images contained in the document one by one. The current image is number {k + 1} out of a total of {len(data)}."
                prompt = Prompt(None, f"{message}", additional_info=additional_info, available_data=[img], history=[str(m) for m in message_history])
                current_answer, current_cost = model.make_prompt_request(prompt)
                answers.append(current_answer)
                cost += current_cost
            answer = merge_answer_pieces(answers)
        else:
            prompt = Prompt(None, f"{message}", available_data=data, history=[str(m) for m in message_history])
            answer, cost = model.make_prompt_request(prompt)
        output.append(Message(answer, [], model.spec.model_id, cost=cost))

    return output



from typing import *
from parsee.datasets.dataset_dataclasses import BaseDatasetRow, Transformation, ColumnSettings


class DatasetWriter:

    def write_rows(self, dataset_rows: List[BaseDatasetRow], dataset: str):
        self.check_consistency(dataset_rows)
        self._write_rows(dataset_rows, dataset)

    def _write_rows(self, dataset_rows: List[BaseDatasetRow], dataset: str):
        raise NotImplementedError

    def check_consistency(self, rows: List[BaseDatasetRow]):
        all_keys = set()
        for row in rows:
            all_keys.update(row.column_names())
        for row in rows:
            current_column_names = set(row.column_names())
            diff = all_keys.difference(current_column_names)
            if len(diff) > 0:
                row.add_missing_columns(list(diff))

    def finish_writing(self):
        return


class ModelWriter:

    local_write_dir: str

    def set_sub_directory(self, sub_directory_name: str):
        raise NotImplementedError

    def save_model_settings(self, columns: Dict[str, ColumnSettings]):
        raise NotImplementedError


from typing import *
from functools import reduce

from parsee.templates.job_template import JobTemplate
from parsee.templates.helpers import create_template, StructuringItem, OutputType
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.storage.interfaces import StorageManager
from parsee.extraction.extractor_elements import StandardDocumentFormat, FinalOutputTableColumn
from parsee.extraction.extractor_dataclasses import ParseeAnswer, ParseeBucket
from parsee.storage.in_memory_storage import InMemoryStorageManager
from parsee.converters.image_creation import ImageCreator
from parsee.extraction.final_structuring import get_structured_tables_from_locations, final_tables_from_columns
from parsee.extraction.models.model_loader import element_models_from_schema, meta_models_from_items, question_models_from_schema, mapping_models_from_schema, ModelLoader


def _model_loader(models: List[MlModelSpecification], custom_model_loader: Optional[ModelLoader] = None, custom_image_creator: Optional[ImageCreator] = None) -> ModelLoader:
    model_loader = custom_model_loader
    if model_loader is None:
        storage = InMemoryStorageManager(models, custom_image_creator)
        model_loader = ModelLoader(storage)
    return model_loader


def run_custom_prompt(doc: StandardDocumentFormat, prompt: str, model: MlModelSpecification, custom_model_loader: Optional[ModelLoader] = None, custom_image_creator: Optional[ImageCreator] = None) -> str:
    """
    Runs a custom prompt and returns the unmodified response from the model.
    Will insert data (either text or images, depending on model type selcted) from your document the same way the extraction via templates works.
    """
    template = create_template([StructuringItem(prompt, OutputType.TEXT)])
    # update the models
    template.set_default_model(model)
    _, _, answers_text = structure_data(doc, template, _model_loader([model], custom_model_loader, custom_image_creator), {})
    if len(answers_text) < 1:
        return ""
    return answers_text[0].raw_value


def run_job_with_single_model(doc: StandardDocumentFormat, job_template: JobTemplate, model: MlModelSpecification, custom_model_loader: Optional[ModelLoader] = None, custom_image_creator: Optional[ImageCreator] = None) -> Tuple[List[ParseeBucket], List[FinalOutputTableColumn], List[ParseeAnswer]]:
    model_loader = _model_loader([model], custom_model_loader, custom_image_creator)
    # update the models
    job_template.set_default_model(model)
    return structure_data(doc, job_template, model_loader, {})


def structure_data(doc: StandardDocumentFormat, job_template: JobTemplate, model_loader: ModelLoader, params: Dict[str, Any]) -> Tuple[List[ParseeBucket], List[FinalOutputTableColumn], List[ParseeAnswer]]:

    # add manual answers to params
    params = {**params, **job_template.detection.settings, **job_template.questions.settings}

    question_models = question_models_from_schema(job_template.questions, job_template.meta, model_loader, params)

    answers: List[ParseeAnswer] = []
    for question_model in question_models:
        answers += question_model.predict_answers(doc)

    models_loc = element_models_from_schema(job_template.detection, model_loader, params)

    locations = []
    for model_loc in models_loc:
        locations += model_loc.classify_elements(doc)

    output_values = get_structured_tables_from_locations(job_template, doc, locations)

    # add meta values
    all_meta_ids = list(set(reduce(lambda acc, x: acc+x.metaInfoIds, job_template.detection.items, [])))
    meta_ids_by_main_class = reduce(lambda acc, x: {**acc, x.id: x.metaInfoIds}, job_template.detection.items, {})
    if len(all_meta_ids) > 0:
        meta_models = meta_models_from_items([x for x in job_template.meta if x.id in all_meta_ids], model_loader, params)
        for meta_model in meta_models:
            meta_predictions_list = meta_model.predict_meta(output_values, doc.elements)
            for k, meta_predictions in enumerate(meta_predictions_list):
                output_values[k].meta += [x for x in meta_predictions if x.class_id in meta_ids_by_main_class[output_values[k].detected_class]]

    # run mapping
    all_mappings: List[ParseeBucket] = []
    tables = final_tables_from_columns(output_values)
    mapping_models = mapping_models_from_schema(job_template.detection, model_loader, params)
    for model_mapping in mapping_models:
        for table in tables:
            mappings, mapping_schema = model_mapping.classify_elements(table)
            all_mappings += mappings

    return all_mappings, output_values, answers


import os
import time
from typing import List, Tuple
from dataclasses import dataclass
from decimal import Decimal
import math

import anthropic
import tiktoken

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.extractor_dataclasses import Base64Image


class AnthropicModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.max_retries = 5
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024 if model.max_output_tokens is None else model.max_output_tokens
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer
        self.client = anthropic.Anthropic(api_key=model.api_key if model.api_key is not None else os.getenv("ANTHROPIC_API_KEY"))

    def _call_api(self, prompt: str, images: List[Base64Image]) -> Tuple[str, Decimal]:
        user_message_content = [
            {
                "type": "text",
                "text": prompt
            }
        ]
        if len(images) > 0:
            user_message_content += [
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": x.media_type,
                        "data": x.data,
                    },
                }
                for x in images
            ]

        message = self.client.messages.create(
            model=self.spec.internal_name,
            max_tokens=self.max_tokens_answer,
            temperature=0,
            system=self.spec.system_message if self.spec.system_message is not None else "",
            messages=[
                {"role": "user", "content": user_message_content}
            ]
        )

        answer = message.content[0].text if len(message.content) > 0 else ""
        cost_input = (message.usage.input_tokens * Decimal(self.spec.price_per_1k_tokens/1000)) if self.spec.price_per_1k_tokens is not None else Decimal(0)
        cost_output = (message.usage.output_tokens * Decimal(self.spec.price_per_1k_output_tokens / 1000)) if self.spec.price_per_1k_output_tokens is not None else Decimal(0)
        cost_images = (len(images) * Decimal(self.spec.price_per_image)) if self.spec.price_per_image is not None else Decimal(0)
        final_cost = cost_input + cost_output + cost_images
        return answer, final_cost

    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, _ = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        return self._call_api(final_prompt, prompt.available_data if self.spec.multimodal else [])



from typing import *
from dataclasses import dataclass
from enum import Enum
from decimal import Decimal

import tiktoken
from tiktoken.core import Encoding

from parsee.utils.enums import SearchStrategy, DocumentType
from parsee.extraction.extractor_elements import FileReference


@dataclass
class Message:
    text: str
    references: List[FileReference]
    author: Optional[str] = None
    cost: Optional[Decimal] = None

    def __str__(self):
        if self.author is None:
            return self.text
        return f"{self.author}: {self.text}"

    def __repr__(self):
        return str(self)


@dataclass
class ChatSettings:
    max_el_in_memory: int = 10000
    max_images_to_load_per_doc: int = 30
    min_tokens_for_instructions_and_history = 500
    encoding: Encoding = tiktoken.get_encoding("cl100k_base")



from typing import *
from decimal import Decimal

from parsee.extraction.tasks.mappings.mapping_model import MappingModel, ElementSchema, MappingSchema, ParseeBucket
from parsee.storage.interfaces import StorageManager
from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel
from parsee.extraction.tasks.mappings.features import LLMMappingFeatureBuilder
from parsee.extraction.extractor_elements import FinalOutputTable
from parsee.utils.constants import ID_NOT_AVAILABLE
from parsee.utils.helper import parse_json_dict, parse_int_simple


class MappingModelLLM(MappingModel):

    def __init__(self, items: List[ElementSchema], storage: StorageManager, llm: LLMBaseModel, **kwargs):
        super().__init__(items)
        self.model_name = llm.spec.model_id
        self.llm = llm
        self.storage = storage
        self.prob = 0.8
        self.feature_builder: LLMMappingFeatureBuilder = LLMMappingFeatureBuilder()

    def parse_answer(self, table: FinalOutputTable, answer: str, schema: MappingSchema, li_identifier: str) -> List[ParseeBucket]:
        answer_dict = parse_json_dict(answer)
        if answer_dict is None or not isinstance(answer_dict, Dict):
            return []
        output = []
        handled = set()
        for bucket in schema.buckets:
            if bucket.id in answer_dict and isinstance(answer_dict[bucket.id], List):
                for answer in answer_dict[bucket.id]:
                    line_item_idx = parse_int_simple(answer)
                    if line_item_idx is not None:
                        item = ParseeBucket(schema.id, bucket.id, li_identifier, self.model_name, self.prob, line_item_idx, 0, Decimal(1))
                        output.append(item)
                        handled.add(line_item_idx)
        # handle items that have not been classified
        for line_item_idx, _ in enumerate(table.line_items):
            if line_item_idx not in handled:
                output.append(ParseeBucket(schema.id, ID_NOT_AVAILABLE, li_identifier, self.model_name, self.prob, line_item_idx, 0, Decimal(1)))
        return output

    def classify_with_schema(self, table: FinalOutputTable, schema: MappingSchema) -> List[ParseeBucket]:

        prompt = self.feature_builder.make_prompt(table, schema)
        answer, amount = self.llm.make_prompt_request(prompt)
        self.storage.log_expense(self.llm.spec.model_id, amount, f"mapping:{table.detected_class}")
        return self.parse_answer(table, answer, schema, table.li_identifier)



from typing import *
import json
import decimal

from parsee.datasets.dataset_dataclasses import MappingUniqueIdentifier
from parsee.extraction.extractor_elements import FinalOutputTable
from parsee.extraction.extractor_dataclasses import ParseeBucket
from parsee.utils.helper import clean_text_for_word_vectors2
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.templates.mappings import MappingBucket
from parsee.datasets.dataset_dataclasses import DatasetRow
from parsee.extraction.tasks.mappings.mapping_model import MappingSchema
from parsee.utils.sample_items import samples


class MappingFeatureBuilder:

    _memory: Dict[MappingUniqueIdentifier, Dict]

    def __init__(self):
        self._memory = {}
        self.number_replacement = "xnumberx"

    def get_text_all(self, table: FinalOutputTable, schema_id: str) -> str:

        items = []
        for kv_index, item in enumerate(table.line_items):
            items.append(self._make_base_features(table, schema_id, kv_index)["caption_cleaned"])

        return ", ".join(items)

    def _get_cleaned_caption(self, table: FinalOutputTable, schema_id: str, kv_index: int) -> str:
        if kv_index < 0 or kv_index > len(table.line_items)-1:
            return ""
        features = self._make_base_features(table, schema_id, kv_index)
        return features["caption_cleaned"]

    def _make_base_features(self, table: FinalOutputTable, schema_id: str, kv_index: int) -> Dict[str, str]:

        unique_identifier = MappingUniqueIdentifier(schema_id, table.li_identifier, kv_index)

        if unique_identifier in self._memory:
            return self._memory[unique_identifier]

        line_item_org = table.line_items[kv_index]
        total_items = len(table.line_items)

        self._memory[unique_identifier] = {"class_id": table.detected_class, "caption_org": line_item_org, "caption_cleaned": clean_text_for_word_vectors2(line_item_org, None, True, True, self.number_replacement), "item_idx": kv_index, "total_items": total_items}

        return self._memory[unique_identifier]

    def make_features(self, source_identifier: Optional[str], template_id: Optional[str], table: FinalOutputTable, schema_id: str, kv_index: int) -> DatasetRow:

        base_features = self._make_base_features(table, schema_id, kv_index)
        unique_identifier = MappingUniqueIdentifier(schema_id, table.li_identifier, kv_index)

        all_text = self.get_text_all(table, schema_id)

        item_before = self._get_cleaned_caption(table, schema_id, kv_index-1)
        item_after = self._get_cleaned_caption(table, schema_id, kv_index+1)

        features = {**base_features, "all_items": all_text, "item_before": item_before, "item_after": item_after}

        return DatasetRow(source_identifier, template_id, unique_identifier, features)


class LLMMappingFeatureBuilder(MappingFeatureBuilder):

    def __init__(self):
        super().__init__()
        self.number_replacement = "[number]"

    def build_raw_answer(self, choices: List[ParseeBucket], schema: MappingSchema) -> str:
        output = {}
        for bucket in schema.buckets:
            output[bucket.id] = [self.line_item_formatter(x.kv_index) for x in choices if x.bucket_id == bucket.id]
        return json.dumps(output)

    def item_id_string(self, bucket: MappingBucket):
        return f"{bucket.id}"

    def format_table(self, table: FinalOutputTable) -> str:
        output = []
        for k, line_item in enumerate(table.line_items):
            row = [f"{self.line_item_formatter(k)}: {line_item}"]
            for col in table.columns:
                row.append(str(col.key_value_pairs[k][1]))
            output.append(row)
        return json.dumps(output)

    def line_item_formatter(self, line_item_idx: int) -> str:
        return f"LI{line_item_idx}"

    def full_example(self) -> str:
        sample_table = samples.table
        schema = samples.schema
        output = {}
        if len(sample_table.line_items) != 3 or len(schema.buckets) != 3:
            raise Exception("needs 3 items")
        output[schema.buckets[0].id] = [self.line_item_formatter(0), self.line_item_formatter(1)]
        output[schema.buckets[1].id] = [self.line_item_formatter(2)],
        output[schema.buckets[2].id] = []
        return json.dumps(output)

    def format_schema(self, schema: MappingSchema):
        bucket_str = ""
        for schema_item in schema.buckets:
            bucket_str += f"{self.item_id_string(schema_item)}: {schema_item.caption}" + (f" ({schema_item.description})"if schema_item.description is not None and schema_item.description.strip() != "" else "") +"\n"
        return bucket_str

    def build_prompt(self, table: FinalOutputTable, schema: MappingSchema) -> Prompt:

        return Prompt(
            "You are supposed to classify line-items from a table into several possible buckets. Each item can only be placed in one bucket. If possible, make sure that buckets that represent a sum of some other items are roughly adding up, using the numbers in the table.",
            f"The table with data will be provided as an array of rows, each line item has an ID that will be included in the first cell at the beginning. After the first cell, all other cells are the values of the line item for each column. The buckets are presented with an ID followed by a descriptive name.", "Format your output as a JSON dictionary, where the keys correspond to the IDs of the buckets and the values are arrays of the line item IDs.",
            f"For example for the following table: {self.format_table(samples.table)} \n"
            f"And the following buckets: {self.format_schema(samples.schema)}"
            f"a valid output would be: {self.full_example()}",
            f"The actual available buckets are the following: {self.format_schema(schema)}\n"
            f"The actual table with line items to be classified is the following: {self.format_table(table)}"
        )

    def make_prompt(self, table: FinalOutputTable, schema: MappingSchema) -> Prompt:

        return self.build_prompt(table, schema)


import os
import time
from typing import List, Tuple
from dataclasses import dataclass
from decimal import Decimal
import math

import openai
import tiktoken

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, get_tokens_encoded, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt
from parsee.extraction.extractor_dataclasses import Base64Image


class ChatGPTModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.max_retries = 5
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024 if model.max_output_tokens is None else model.max_output_tokens
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer
        openai.api_key = model.api_key if model.api_key is not None else os.getenv("OPENAI_KEY")

    def _call_api(self, prompt: str, images: List[Base64Image]) -> Tuple[str, Decimal]:
        user_message_content = [
            {
                "type": "text",
                "text": prompt
            }
        ]
        if len(images) > 0:
            user_message_content += [
                {
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:{x.media_type};base64,{x.data}"
                    }
                }
                for x in images
            ]

        messages = [
                {"role": "user", "content": user_message_content}
            ]

        if self.spec.system_message is not None:
            messages.insert(0, {"role": "system", "content": self.spec.system_message})

        response = openai.ChatCompletion.create(
            model=self.spec.internal_name,
            messages=messages,
            temperature=0,
            max_tokens=self.max_tokens_answer,
            top_p=1,
            frequency_penalty=0,
            presence_penalty=0
        )

        answer = response['choices'][0]["message"]["content"]
        cost_input = (int(response["usage"]['prompt_tokens']) * Decimal(self.spec.price_per_1k_tokens / 1000)) if self.spec.price_per_1k_tokens is not None else Decimal(0)
        cost_output = (int(response["usage"]['completion_tokens']) * Decimal(self.spec.price_per_1k_output_tokens / 1000)) if self.spec.price_per_1k_output_tokens is not None else Decimal(0)
        cost_images = (len(images) * Decimal(self.spec.price_per_image)) if self.spec.price_per_image is not None else Decimal(0)
        final_cost = cost_input + cost_output + cost_images
        return answer, final_cost

    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, _ = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        return self._call_api(final_prompt, prompt.available_data if self.spec.multimodal else [])



import os
import time
from typing import List, Tuple
from dataclasses import dataclass
from decimal import Decimal
import math

from together import Together
import tiktoken

from parsee.extraction.models.llm_models.llm_base_model import LLMBaseModel, get_tokens_encoded, truncate_prompt
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.extraction.models.llm_models.prompts import Prompt


class TogetherModel(LLMBaseModel):

    def __init__(self, model: MlModelSpecification):
        super().__init__(model)
        self.encoding = tiktoken.get_encoding("cl100k_base")
        self.max_tokens_answer = 1024 if model.max_output_tokens is None else model.max_output_tokens
        self.max_tokens_question = self.spec.max_tokens - self.max_tokens_answer
        self.client = Together(api_key=model.api_key if model.api_key is not None else os.getenv("TOGETHER_API_KEY"), max_retries=5)

    def _call_api(self, prompt: str) -> Tuple[str, Decimal]:
        messages = [{"role": "user", "content": prompt}]
        if self.spec.system_message is not None:
            messages.insert(0, {"role": "system", "content": self.spec.system_message})
        response = self.client.chat.completions.create(
            model=self.spec.internal_name,
            messages=messages,
            temperature=0,
            max_tokens=self.max_tokens_answer,
            top_p=1,
        )
        answer = response.choices[0].message.content
        cost_input = (int(response.usage.prompt_tokens) * Decimal(self.spec.price_per_1k_tokens / 1000)) if self.spec.price_per_1k_tokens is not None else Decimal(0)
        cost_output = (int(response.usage.completion_tokens) * Decimal(self.spec.price_per_1k_output_tokens / 1000)) if self.spec.price_per_1k_output_tokens is not None else Decimal(0)
        final_cost = cost_input + cost_output
        return answer, final_cost

    def make_prompt_request(self, prompt: Prompt) -> Tuple[str, Decimal]:
        final_prompt, _ = truncate_prompt(prompt, self.encoding, self.max_tokens_question)
        return self._call_api(final_prompt)



from decimal import Decimal
from typing import *
import math
from functools import reduce

from parsee.templates.job_template import JobTemplate
from parsee.extraction.models.model_dataclasses import MlModelSpecification
from parsee.storage.vector_stores.interfaces import VectorStore
from parsee.extraction.extractor_elements import FileReference
from parsee.converters.image_creation import ImageCreator
from parsee.extraction.extractor_dataclasses import Base64Image
from parsee.chat.custom_dataclasses import ChatSettings


class StorageManager:

    vector_store: VectorStore
    image_creator: ImageCreator

    def __init__(self, vector_store: VectorStore, image_creator: ImageCreator):
        self.vector_store = vector_store
        self.image_creator = image_creator

    def db_values_template(self, job_template: JobTemplate, strict: bool) -> JobTemplate:
        raise NotImplementedError

    def log_expense(self, service: str, amount: Decimal, class_id: str):
        raise NotImplementedError

    def get_available_models(self) -> List[MlModelSpecification]:
        raise NotImplementedError


class DocumentManager:

    storage: StorageManager

    def __init__(self, storage: StorageManager, settings: ChatSettings):
        self.storage = storage
        self.settings = settings

    def _load_documents(self, references: List[FileReference], multimodal: bool, search_term: Optional[str], max_images: Optional[int], max_tokens: Optional[int], load_function: Callable, show_chunk_index: bool) -> Union[str, List[Base64Image]]:
        # find and load the most relevant documents
        docs = []
        unique_identifiers = set([x.source_identifier for x in references])
        for source_identifier in self.storage.vector_store.sort_identifiers_by_relevance(unique_identifiers, search_term):
            total_added = 0
            doc = load_function(source_identifier)
            # check if all elements should be taken or not
            take_all = len([x for x in references if x.source_identifier == doc.source_identifier and x.element_index is None]) > 0
            if not take_all:
                allowed_element_indexes = [x.element_index for x in references if x.source_identifier == doc.source_identifier and x.element_index is not None]
                doc.elements = [x for x in doc.elements if x.source.element_index in allowed_element_indexes]
            if total_added + len(doc.elements) > self.settings.max_el_in_memory:
                to_add = self.settings.max_el_in_memory - total_added
                if to_add > 0:
                    doc.elements = doc.elements[0:to_add]
                else:
                    break
            docs.append(doc)

        if multimodal:
            if show_chunk_index:
                raise Exception("chunks can't be displayed with multimodal option")
            output_by_doc = {}
            total_images = 0
            for doc in docs:
                output_by_doc[doc.source_identifier] = self.storage.image_creator.get_images(doc, doc.elements, self.settings.max_images_to_load_per_doc, None)
                total_images += len(output_by_doc[doc.source_identifier])
            if max_images is not None and total_images > max_images:
                max_images_per_file = math.floor(max_images / len(output_by_doc.keys()))
                output = []
                for k, values in output_by_doc.items():
                    if len(values) > max_images_per_file:
                        if k == list(output_by_doc.keys())[-1]:
                            images_left = max_images - len(output)
                            output += values[0:images_left]
                        else:
                            output += values[0:max_images_per_file]
                    else:
                        output += values
                return output
            else:
                return reduce(lambda acc, x: acc + x, output_by_doc.values(), [])
        else:
            output_by_doc = {}
            total_tokens = 0
            for k, doc in enumerate(docs):
                output_by_doc[doc.source_identifier] = self.settings.encoding.encode(doc.to_string(show_chunk_index))
                total_tokens += len(output_by_doc[doc.source_identifier])
            if total_tokens > max_tokens:
                max_tokens_per_document = math.floor(max_tokens / len(output_by_doc.keys()))
                output = ""
                for k, tokens in enumerate(output_by_doc.values()):
                    output += f"[START OF DOCUMENT with index {k}]\n"
                    output += self.settings.encoding.decode(tokens[0:max_tokens_per_document])
                    output += f"[END OF DOCUMENT with index {k}]\n\n"
                return output
            else:
                output = ""
                for k, doc in enumerate(docs):
                    output += f"[START OF DOCUMENT with index {k}]\n"
                    output += doc.to_string(show_chunk_index)
                    output += f"[END OF DOCUMENT with index {k}]\n\n"
                return output

    def load_documents(self, references: List[FileReference], multimodal: bool, search_term: Optional[str], max_images: Optional[int], max_tokens: Optional[int], show_chunk_index: bool = False) -> Union[str, List[Base64Image]]:
        raise NotImplementedError
